#=======================================================================
#	$Id: DBHash.pm,v 1.5 2005/09/09 11:05:09 pythontech Exp pythontech $
#	Triple Store using Berkeley DB 'Hash'
#	Copyright (C) 2004,2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples::DBHash;
use strict;
use PythonTech::Triples::TiedHash;
use DB_File;
use IO::File;
use Fcntl qw(:flock);

our @ISA = qw(PythonTech::Triples::TiedHash);

#-----------------------------------------------------------------------
#	Create new triplestore from Berkeley DB file
#	  filename	Name of Berkeley DB file
#	  access	Access mode letters: w=write, l=lock
#	Inherited from TiedHash:
#	  index_s2p	Request extra index: all pred for each subj
#	  index_o2p	Request extra index: all pred for each obj
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $props = ref($_[0]) eq 'HASH' ? {%{$_[0]}} : {@_};
    my $dbfilename = delete $props->{'filename'};
    my $access = delete $props->{'access'};
    my $self = {
	dbfilename => $dbfilename,
	access => $access,
	db => undef,
	dbtie => undef,
	lockhandle => undef,
    };
    bless $self, $class;

    if (defined $dbfilename) {
	$self->reconnect;	# First connection
    }
    # Let superclass inspect optional features
    $self->_update_features;	# Check for p,P flags
    # Let superclass deal with any other properties
    $self->configure($props);
    return $self;
}

sub commit {
    my($self) = @_;
#    print STDERR (map {"$_ => $self->{'mods'}->{$_}\n"} keys %{$self->{'mods'}}),"\n";
    $self->SUPER::commit();
    if ($self->{'dbtie'}) {
#	print STDERR "tied=",tied(%{$self->{'db'}});
#	print STDERR "$$ syncing\n";
	my $status = $self->{'dbtie'}->sync;
#	print STDERR "$$ sync status: $status\n";
    }
}

sub disconnect {
    my($self) = @_;
    untie %{$self->{'db'}};
    $self->{'dbtie'} = undef;
    if ($self->{'access'} =~ /l/) {
	flock($self->{'lockhandle'}, LOCK_UN) or
	    die "Error unlocking $self->{'dbfilename'}: $!\n";
	$self->{'lockhandle'}->close or
	    die "Error closing $self->{'dbfilename'}: $!\n";
	$self->{'lockhandle'} = undef;
    }
}

#-----------------------------------------------------------------------
#	Reconnect to backend db file, possibly with different access
#-----------------------------------------------------------------------
sub reconnect {
    my($self, $access) = @_;

    # Ensure disconnected
    $self->disconnect if $self->{'dbtie'};
    
    # Maybe change access mode
    $self->{'access'} = $access
	if defined $access;

    my $dbfilename = $self->{'dbfilename'};
#    print STDERR "$self reconnect: dbfilename=$dbfilename\n";
    my %db;
    my $tie;
    my $lockhandle;
    my $writing = $self->{'access'} =~ /w/;
    my $mode = $writing ? (O_RDWR|O_CREAT) : O_RDONLY;
    if ($self->{'access'} =~ /l/) {
	# Lock it using an extra file handle
	if ($writing) {
	    my $lockmode = (-f $dbfilename ? "+<" : ">");
	    $lockhandle = IO::File->new($dbfilename, $lockmode) or
		die "Cannot open $dbfilename with IO::File: $!\n";
	    # Need exclusive access for writing
#	    print STDERR "$$ locking LOCK_EX\n";
	    flock($lockhandle, LOCK_EX) or
		die "Error locking $dbfilename: $!\n";
	} else {
	    # Read-only, so can use non-exclusive lock.
	    # Allows other readers, but no writers.
	    $lockhandle = IO::File->new($dbfilename, "<") or
		die "Cannot open $dbfilename with IO::File: $!\n";
	    # Shared access is OK for reading
	    flock($lockhandle, LOCK_SH) or
		die "Error locking $dbfilename: $!\n";
	}
    }
#	print STDERR "$$ tying\n";
    $tie = tie %db, 'DB_File', $dbfilename, $mode, 0644
	or die "Cannot tie $dbfilename: $!\n";

    $self->{'db'} = \%db;
    $self->{'dbtie'} = $tie;
    $self->{'lockhandle'} = $lockhandle;
}

1;
